#!/usr/bin/perl
use Win32::Console;
###############################################################################
###############################################################################
###
### \\ Written by -=Assarbad=- <oliver at assarbad dot net>          2004-09-30
### \\ Last modification: 2004-10-08
### \\
### \\ This code is published under the terms of the Mozilla Public License
### \\ (MPL) to be compatible with the project JEDI:
### \\
### \\                        http://www.delphi-jedi.org
###
### Small Perl script to translate Microsoft IDL files to Object Pascal/Delphi
### units.
###
### Requires Windows and several modules from ActiveState's Perl distro to run.
###
###############################################################################
###############################################################################

############################ Variables
my ($p1, $outfile) = @ARGV; # Get the argument holding the IDL file name
my $ARGC = @ARGV; # number of arguments
my $STDOUT = new Win32::Console(STD_OUTPUT_HANDLE); # Get the STDOUT "handle"
my $OldAttr = $STDOUT->Attr(); # Save the old colors
my %Friends = (
   'REFIID' => 'PIID',
   'LPSYSTEMTIME' => 'PSystemTime',
   'REFCLSID' => 'PCLSID'); # Lookup of type names ...
############################ Info about script
$STDOUT->Cls();
$STDOUT->Attr($FG_LIGHTGREEN | $BG_BLACK);
print "the fIDLer - IDL parser/converter 0.1\n";
print "For more information use parameters --help or --copyleft\n\n";
$STDOUT->Attr($OldAttr);
############################ Check the input
if ($ARGC == 0) # no command line parameters given
{
  $STDOUT->Attr($FG_LIGHTRED | $BG_BLACK); # Light red on black
  print "Error: Wrong Syntax!\n\n"; # Error
  $STDOUT->Attr($OldAttr); # Restore color
  # Print correct syntax
  my $x = $0; $x =~ s/^.+\\([^\\]+)$/$1/;
  print "Usage: $x <filename>\n\n";
  print "   filename = Name of an IDL file. The .IDL extension is optional.\n";
}
else
{
  if($p1 =~ /^--help$/i)
  {
    PrintHelp($STDOUT);
  }
    elsif($p1 =~ /^--copyleft$/i)
    {
      PrintCopyrightInfo($STDOUT);
    }
      else
      {
      	 ParseIDL($STDOUT, $p1);
      }
}

############################ Bail out
$STDOUT->Attr($OldAttr); # Restore old colors
exit;

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Functions (subs) for this script
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

sub PrintHelp
#------------------------------------------------------------------------------
#  Print the help text contain below the __END__ tag in this script
#------------------------------------------------------------------------------
{
  my ($STDOUT) = @_;
  print join('', <DATA>);
}

#------------------------------------------------------------------------------
#  Parse the script for copyright information
#------------------------------------------------------------------------------
sub PrintCopyrightInfo
{
  my ($STDOUT) = @_;
  $STDOUT->Attr($FG_WHITE | $BG_BLACK);
  open (FILE, '<'.$0); # Read this script file
  foreach(<FILE>)
  {
    # Only output files starting with ### \\
    if ($_ =~ /^### \\\\?[\t\s]*(.*)/i)
    {
      print $1."\n";
    }
  }
  close(FILE);
}

#------------------------------------------------------------------------------
#  Parse the IDL content
#------------------------------------------------------------------------------
sub ParseIDL
{
  # Get the function parameters
  my ($STDOUT, $fname) = @_;
  # Read the input file
  open (FILE, $fname) or die("Could not open \"$fname\"!");
  my @fconta = <FILE>;
  my @output; # For output
  close(FILE);
  # Parse all lines
  for($i = 0; $i < scalar @fconta; $i++)
  {
    my $elem = $fconta[$i];
#--------------------------------------
# Handle empty lines
#--------------------------------------
    if ($elem =~ /^[\s\t\n]*$/) # skip empty line (includes space-only lines
    {
      push(@output, "\n"); # Output empty line
      next;
    }
#--------------------------------------
# Handle single-line comments
#--------------------------------------
    if ($elem =~ /^[\t\s]*\/\//) # skip single-line comments
    {
      push(@output, $elem); # Output commented line
      next;
    }
#--------------------------------------
# Handle CPP_QUOTEs, assume single-line
#--------------------------------------
    if ($elem =~ /^[\t\s]*cpp_quote\("(.+)"\)/i)
    {
      Treat_C_Quotes($1, $elem, \@output);
      next;
    }
#--------------------------------------
# Handle import, assume single-line
#--------------------------------------
    if ($elem =~ /^[\t\s]*import/i)
    {
      push(@output, "\n// \@REVIEW\@ (require additional unit in USES?): ".$elem);
      next;
    }
#--------------------------------------
# Handle interface forward decl.
#--------------------------------------
    if ($elem =~ /^[\t\s]*interface[\t\s]+(.+?);/i)
    {
      push(@output, "\n// ".$elem);
      next;
    }
#--------------------------------------
# Handle C++ consts, single-line
#--------------------------------------
    if ($elem =~ /^[\t\s]*const[\t\s]+(.+?)[\t\s]+(.+?)[\t\s]*=[\t\s]*(.+?);/i) # Check syntax
    {
      my ($type, $name, $value) = ($1, $2, $3);
      if (lc($type) eq 'int') {$type = 'Integer'};
      if (lc($type) eq 'float') {$type = 'Real'};
      push(@output, "const\n  {\$EXTERNALSYM ".VarName($name)."}\n  ".VarName($name)." : $type = $value;\n");
      next;
    }
#--------------------------------------
# Handle the rest (multi-line expr.)
#--------------------------------------
    # We can conclude, that at this point there are no comments or C-quotes or empty lines
    if ($elem =~ /^[\t\s]*(\[|struct|typedef)/i) # We don't know anything else, yet.
    {
      if (lc($1) eq 'struct')
      {
        # Iterate through the array and append if no match!
        while ($i < scalar @fconta)
      	{
          # Try to match
          unless ($elem =~ /^[\t\s\n]*struct[\t\s\n]+(.+?)[\t\s\n]*;/gism)
            {
              $elem .= $fconta[++$i];
              next; # while
            }
          else
            {
              push(@output, "type\n  {\$EXTERNALSYM $1}\n  $1 = record end;\n\n");
              last;
            }
      	}
      	next; # Next in the FOR loop
      }
        elsif ($elem =~ /^[\t\s]*\[[\t\s]*\n/i)
        {
          # Read the interface attributes
          while($i < scalar @fconta) {$elem .= $fconta[++$i]; if($elem =~ /[\t\s]*\][\t\s]*\n$/gism) {last;}}
          my %data = ();
          # Extract the GUID
          if ($elem =~ /uuid\(([a-f0-9]+-[a-f0-9]+-[a-f0-9]+-[a-f0-9]+-[a-f0-9]+)\)/i)
            {$data{GUID} = "['{".uc($1)."}']"};
          # ... the rest of the attributes being ignored
          $elem = $fconta[++$i];
          # Read the actual interface declaration
          if ($elem =~ /^[\t\s]*interface[\t\s\n]+(.+?)[\t\s\n]*:[\t\s\n]*(.+?)\n/gism)
          {
            $data{name} = $1; $data{descend} = $2;
            while($i < scalar @fconta) {$elem .= $fconta[++$i]; if($elem =~ /^[\t\s]*\}[\t\s]*[;]*[\t\s]*\n$/gism) {last;}}
            # At this point a complete interface declaration should be in $elem
            push(@output, Treat_Interface($elem, \@output, \%data));
          }
          next;
        }
          elsif ($elem =~ /^[\t\s]*typedef[\t\s\n]+enum/gism)
          {
            while($i < scalar @fconta) {$elem .= $fconta[++$i]; if($elem =~ /\}(.+?);\n$/gism) {last;}}
            push(@output, Treat_Enum($elem));
            next;
          }
            elsif ($elem =~ /^[\t\s]*typedef[\t\s\n]+struct[\t\s\n]+[^\{\};]+;/gism)
            {
              push(@output, Treat_Struct_NoCurl($elem));
              next;
            }
              elsif ($elem =~ /^[\t\s]*typedef[\t\s\n]+struct/gism)
              {
                while($i < scalar @fconta) {$elem .= $fconta[++$i]; if($elem =~ /\}(.+?);\n$/gism) {last;}}
                push(@output, Treat_Struct_Curl($elem));
                next;
              }
                elsif ($elem =~ /^[\t\s]*typedef[\t\s\n]+union/gism)
                {
                  while($i < scalar @fconta) {$elem .= $fconta[++$i]; if($elem =~ /\}(.+?);\n$/gism) {last;}}
                  push(@output, Treat_Union($elem));
                  next;
                }
    }
#--------------------------------------
# Handle errors
#--------------------------------------
    push(@output, "// @REVIEW@ (completely unknown syntax element): $elem");
    next;
  }
  OutputResult($fname, \@output);
}

#------------------------------------------------------------------------------
#  Treat lines with cpp_quote()
#------------------------------------------------------------------------------
sub Treat_C_Quotes
{
  my ($temp, $full, $arr) = @_;
  my $match = 0; # Number of matches below - to allow detection of non-matched lines
  # Replace for constant hex numbers
  if($temp =~ s/#define[\t\s]+(.+?)[\t\s]+\(0x([a-f0-9]+?)\)/const\n  \{\$EXTERNALSYM $1\}\n  $1 = \$$2;/i)
    {$match++;}
  # Replace for constant dec numbers
  if($temp =~ s/#define[\t\s]+(.+?)[\t\s]+\(([0-9]+?)\)/const\n  \{\$EXTERNALSYM $1\}\n  $1 = $2;/i)
    {$match++;}
  # For DEFINE_GUID
  if($temp =~ s/DEFINE_GUID\((CLSID|IID)_(.+?),[\t\s]*(.+?),[\t\s]*(.+?),[\t\s]*(.+?),[\t\s]*(.+?)\);/const\n  \{\$EXTERNALSYM $1_$2\}\n  $1_$2 : T$1 = (D1 : $3; D2 : $4; D3 : $5; D4 : ($6));/i)
    {$temp =~ s/0x([0-9a-f]+)[L]*/\$$1/gism; $match++} # Replace 0x by $ and strip L's for long
  # For EXTERN_C
  if($temp =~ s/EXTERN_C/\/\/ EXTERN_C/i)
    {$match++;}
  # For quoted single-line comments
  if ($temp =~ /^[\t\s]*\/\//i)
    {$match++;}
  # For empty lines
  if ($temp =~ /^[\s\t]*$/i)
    {$match++;}
  if ($match > 0)
  {
    push(@{$arr}, "$temp\n"); # Output commented line
  }
  else
  {
    push(@{$arr}, "\n// \@REVIEW\@ (embedded C): ".$full); # Output the original line tagged as error
    if($temp =~ /^[\t\s]*#define[\t\s]*([^\s]+?)[\t\s]*([^\s]+?)$/i)
    {
      # If it is a "#define abc def" then comment it
      $temp = "(*\n// Proposal:\nconst\n  {\$EXTERNALSYM $1}\n  $1 = $2;\n// Instead of \"const\", \"type\" may be appropriate or even advanced treatment\n*)";
      push(@{$arr}, "$temp\n"); # Output commented line
    }
  }
}

#------------------------------------------------------------------------------
#  Parse enum definition
#------------------------------------------------------------------------------
sub Treat_Enum
{
  my($elem) = @_;
  my $result = '';

  if($elem =~ /^[\t\s]*typedef[\t\s\n]+enum[\t\s\n]+(.+?)[\t\s\n]*\{(.+?)\}(.+);[\t\s\n]+$/gism)
  {
    my ($name, $content, $other) = ($1, $2, $3);
    $content =~ s/=[\t\s]*([0-9]|0x[0-9a-f]?)([\t\s]*[,]{0,1})(.*?)\n/$2 \/\/ $1 $3\n/gism;
    $content = InsertExternalSyms($content);
    $result = "  {\$EXTERNALSYM ".VarName($name)."}\n  ".VarName($name)." = (\n$content\n  );".GetOtherNames(VarName($name), $other);
  }
  $result = "type\n"."// \@REVIEW\@ (careful, C-enums can start with indices >0 or have gaps)!\n".$result;
  return "$result\n";
}

#------------------------------------------------------------------------------
#  Insert the {EXTERNALSYM xxx} directives ;)
#------------------------------------------------------------------------------
sub InsertExternalSyms
{
  my ($content) = @_;
  my $ret = '';

  @vars = split(/\n/, $content);
  map(s/\n//i, @vars); # Remove line breaks

  foreach(@vars)
  {
    unless ($_ =~ /^[\t\s]*$/)
    {
      $_ =~ s/([\s\t]*)([^\s\t]+?)([\s\t]+)([,]*)(.+)$/$1\{\$EXTERNALSYM $2\}\n$1$2$3$4$5\n/i;
#      print "$_---------------------------\n";
      $ret .= $_;
    }
  }
  return $ret;
}

#------------------------------------------------------------------------------
#  Parse a struct "assignment" - not real record.
#------------------------------------------------------------------------------
sub Treat_Struct_NoCurl
{
  my($elem) = @_;
  my $result = '';

  if ($elem =~ /^[\t\s]*typedef[\t\s\n]+struct[\t\s\n]+(.+?)[\t\s\n]+([^\{\};]+);/gism)
  {
    my ($name, $newtype) = ($1, $2);
    if ($newtype =~ s/[\t\s\n]*\*[\t\s\n]*(.+)/$1/i)
    {
      $type = "\^$type";
    }
    $result = "  {\$EXTERNALSYM $newtype}\n  $newtype = $type;";
  }
  $result = "type\n"."// \@REVIEW\@ (review this. Could have been ambigous)!\n".$result;
  return "$result\n";
}

#------------------------------------------------------------------------------
#  Treat a real record with members
#------------------------------------------------------------------------------
sub Treat_Struct_Curl
{
  my($elem) = @_;
  my $result = '';

  if ($elem =~ /^[\t\s]*typedef[\t\s\n]+struct[\t\s\n]+(.+?)[\t\s\n]*\{(.+?)\}(.+?);\n$/gism)
  {
    my ($name, $content, $other) = ($1, $2, $3);
    $content =~ s/([0-9a-z_]+)[\t\s]*([0-9a-z_]+);/$2 : $1;/gism;
    $result = "  {\$EXTERNALSYM ".VarName($name)."}\n  ".VarName($name)." = packed record$content\n  end;".GetOtherNames(VarName($name), $other);
  }
  $result = "type\n"."// \@REVIEW\@ (review this. Could have been ambigous)!\n".$result;
  return "$result\n";
}

#------------------------------------------------------------------------------
#  Treat union
#------------------------------------------------------------------------------
sub Treat_Union
{
  my($elem) = @_;
  my $result = '';

  if ($elem =~ /^[\t\s]*typedef[\t\s\n]+union[\t\s\n]+(.+?)[\t\s\n]*\{[\n]*(.+?)\}(.+?);\n$/gism)
  {
    my ($name, $content, $other) = ($1, $2, $3);
    $content =~ s/([0-9a-z_]+)[\t\s]*([0-9a-z_]+);/\( $2 : $1 \);/gism;
    $result = "  {\$EXTERNALSYM ".VarName($name)."}\n  ".VarName($name)." = packed record\n    case Integer of\n";
    my @x = split(/\n/, $content);
    my $i = 0;
    foreach(@x)
      {$result .= "    ".$i++." : $_\n";}
    $result .= "  end;".GetOtherNames(VarName($name), $other);
  }
  $result = "type\n"."// \@REVIEW\@ (review this. Could have been ambigous)!\n".$result;
  return "$result\n";
}

#------------------------------------------------------------------------------
#  Treat the interface declarations
#------------------------------------------------------------------------------
sub Treat_Interface
{
  my ($elem, $arr, $hash) = @_;
  my $result = '';

  # Match format and extract parts ...
  if ($elem =~ /^[\t\s]*interface[\t\s\n]+(.+?)[\t\s\n]*:[\t\s\n]*(.+?)\n[\t\s\n]*\{[\t\s\n]*(.+?)[\t\s]*\}[\t\s]*[;]*[\t\s]*\n$/gism)
  {
    my $contents = $3;
    $Friends{$name} = $name;
    my @methods = split(/\);\n/, $contents);
    map($_ .= ");", @methods);
    map($_ =~ s/\[in,(\s*)out\]/\[in;$1out\]/gism, @methods); # comment out the "attributes"
    map($_ =~ s/\[(in|out|in;\s*out)\]/\{$1\}/gism, @methods); # comment out the "attributes"
    map($_ =~ s/\([\t\s\n]*void[\t\s\n]*\)/\(\)/gsm, @methods); # void is unknown in Delphi
    map($_ =~ s/\n[\t\s]*\n/\n/gism, @methods); # Double \n
    map($_ =~ s/^\n//gism, @methods); # Method records starting with \n
    foreach(@methods)
    {
      $result .= DecideFuncProc($_);
    }
  }
  $result = "type\n  {\$EXTERNALSYM ${$hash}{name}}\n  ${$hash}{name} = interface(${$hash}{descend})\n    ${$hash}{GUID}\n".$result."  end;\n";

  return $result;
}

#------------------------------------------------------------------------------
#  Decide wether the method should be function or procedure and order params
#------------------------------------------------------------------------------
sub DecideFuncProc
{
  my ($record) = @_;
  my $result = '';

  if ($record =~ /^[\t\s]*\/\/\s*(.+?)\n/gism)
  {
    $result .= "// $1\n"; # Make comment
    $record =~ s/^[\t\s]*\/\/\s*(.+?)\n//gism; # Remove the match
  }

  # Match the pattern wether procedure or function
  if ($record =~ /^[\t\s]*([a-z0-9_]+?)[\t\s]+([a-z0-9_]+?)[\t\s]*\((.*?)\)/gis)
  {
    my ($name, $returns, $other) = ($2, $1, $3);
    $result .= ConvertArgs($name, $returns, $other);
  }
    elsif ($record =~ /^[\t\s]*([a-z0-9_]+?)[\t\s]*\((.*?)\)/gis)
    {
      my ($name, $other) = ($1, $2);
      $result .= ConvertArgs($name, '', $other);
    }
      else
      {
        $result .= "// \@REVIEW\@ (malformed method decl.?):\n(*\n$record\n*)\n";
      }

  return $result;
}

#------------------------------------------------------------------------------
#  Convert name, arguments and return to some meaningful Delphi expression
#------------------------------------------------------------------------------
sub ConvertArgs
{
  my ($name, $returns, $other) = @_;
  my $dlpother = '';

  if ($returns =~ /void/) {$returns ='';}
  my $proctype = ($returns eq '') ? "procedure" : "function";
  if ($other eq '')
  {
    return sprintf("    %s %s()%s; stdcall;\n", $proctype, VarName($name), ($returns eq '') ? '' : ": $returns");
  }
  $other =~ s/\n//gism;
  $other =~ s/([\t]+|[\s]{2,})/ /gism; # Remove tabs and double spaces!
  $other =~ s/^\s//gism;
  my @x = split(/,/, $other);
  foreach(@x)
  {
    $dlpother .= GetDelphiArg($_);
  }
  $dlpother =~ s/;\s*$//i;
  return sprintf("    %s %s(%s)%s; stdcall;\n    (*| Parameter(s) was/were [CPP]: %s |*)\n", $proctype, VarName($name), $dlpother, ($returns eq '') ? '' : ": $returns", $other);
}

#------------------------------------------------------------------------------
#  Splits the C argument and converts it by calling a helper function
#------------------------------------------------------------------------------
sub GetDelphiArg
{
  my ($other) = @_;
  my $result = '';

  if ($other =~ /(\{(in|out|in,\s*out)\}){0,1}\s*((const\s+){0,1}[a-z0-9_]+?)(\s+[\*]+?){0,1}[\s]+([a-z0-9_]+)(\[[0-9]*\]){0,1}/i)
  {
    my ($type, $comment, $pointer, $varname, $ext) = ($3, $1, $5, $6, $7);
    # $3 == type
    # $1 == comment
    # $5 == pointer?
    # $6 == variable name
    $pointer =~ s/\s*([\*]*)\s*/$1/i;
    $type =~ s/\s*(const\s+){0,1}([a-z0-9_]+)\s*/$2/i;
    $result .= sprintf("%s;", GetDelphiType($varname, $type, $pointer, $comment, $ext));
  }
  return $result;
}

#------------------------------------------------------------------------------
#  Helper for the above function. It order the arguments and returns a single
#  Delphi-like expression
#------------------------------------------------------------------------------
sub GetDelphiType
{
  my ($varname, $type, $pointer, $comment, $ext) = @_;
  my $result = '';
  my $vartype = 'var ';
  if ($comment eq '{out}') {$vartype = 'out ';}
  if ($comment eq '{in}') {$vartype = 'const ';}

  if ($pointer eq '')
  {
    $type =~ s/^\s*int\s*$/Integer/i; # Replace
    my $xtype = $Friends{$type};
#    print "xtype == $xtype\n";
    if (substr($xtype,0,1) eq 'P')
    {
#       print "--> catched xtype == $xtype\n";
       $result .= $vartype;
       $type = 'T'.substr($xtype, 1, length($xtype));
       print "new type == $type\n";
    }
    $type =~ s/^SYSTEMTIME$/TSystemTime/i; #Replace
    if ($ext ne '')
    {
      $result .= '{@REVIEW@ (was array!!! should be pointer-type?)} ';
    }
    $result .= "$varname : $type";
  }
    elsif($pointer eq '*')
    {
      $result .= $vartype;
      $result .= "$varname : $type";
    }
      elsif($pointer eq '**')
      {
        if ($type =~ /^BYTE$/i)
        {
          $result .= $vartype.' ';
          $type = 'PByte';
        }
        else
          {
            $result .= $vartype.'{@REVIEW@ (double pointer?)} ';
          }
        $result .= "$varname : $type";
      }

  return $result;
}

#------------------------------------------------------------------------------
#  Get aliases for a type identifier
#------------------------------------------------------------------------------
sub GetOtherNames
{
  my ($name, $other) = @_;
  my $result = '';

  my @x = split(/,/, $other);
  foreach(@x)
  {
    if ($_ =~ s/[\t\s\n]*\*[\t\s\n]*(.+)/$1/gism)
      {$result .= "\n  {\$EXTERNALSYM $1}\n  $1 = ^$name;\n".GetDelphiName($1, $name, 1);}
    elsif ($_ =~ s/[\t\s\n]*(.+)/$1/gism)
      {$result .= "\n  {\$EXTERNALSYM $1}\n  $1 =  $name;\n".GetDelphiName($1, $name, 0);}
  }
  return $result;
}

#------------------------------------------------------------------------------
#  Additionally create Delphi-compliant names
#------------------------------------------------------------------------------
sub GetDelphiName
{
  my ($name, $type, $p) = @_;
  my $result = '';

  $name =~ s/[\s\t]*(.+?)[\s\t]*/$1/i;
  my $temp = $name; # Save it

  if ($p == 1) {$name = substr($name, 1, length($name));}
  my @x = split('_', $name);

  foreach(@x)
    {$result .= ucfirst(lc($_));}
  $result = (($p == 0) ? "T" : "P").$result;
  # Save the name into the lookup table ...
  $Friends{$temp} = $result;

  return "  $result = ".(($p == 0) ? " " : "^")."$type;\n";
}

#------------------------------------------------------------------------------
#  Avoids special names vor record members and parameters
#------------------------------------------------------------------------------
sub VarName
{
  my ($arg) = @_;
  if ($arg =~ /^(type|string|var)$/i)
  {
    return $arg.'_';
  }
  return $arg;
}

#------------------------------------------------------------------------------
#  Output the final result and pre-/append some Delphi-stuff
#------------------------------------------------------------------------------
sub OutputResult
{
  my ($fname, $arr) = @_;
  my $comment = "(*****************************************************************************\n".
                "  This IDL-file has been converted by \"the fIDLer\".\n".
                "  [written by -=Assarbad=- <oliver at assarbad dot net> Sept-2004] under MPL\n".
                "  Visit the fIDLer homepage at: http://assarbad.net/en/stuff/\n".
                "  {The 3 above lines should be retained}\n".
                "\n".
                "  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n".
                "  NOTE:\n\n".
                "  Please look for \@REVIEW\@ tags and verify the correctness of the partial\n".
                "  conversion result! <-- delete these 2 lines when done!\n".
                "  There's no guarantee for correct case of parameter or variable types.\n".
                "  If you have a type like BLA_YADDA in IDL then fIDLer will have converted it\n".
                "  to 'TBlaYadda' already. But if the type identifier was BLAYADDA and both\n".
                "  BLA and YADDA being distinct words the result will not be correctly\n".
                "  capitalized!\n".
                "  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n".
                "\n".
                "  The original file was '".$fname."'\n".
                "  File converted: ".GetCurrentDateTime()."\n".
                " *****************************************************************************)\n\n";
  $fname =~ s/\.idl//i;
  # UNIT HEADER
  my $outstr = $comment;
  $outstr .= "unit $fname;\n\n";
  $outstr .= "{\$ALIGN ON}\n";
  $outstr .= "{\$MINENUMSIZE 4}\n";
  $outstr .= "{\$WEAKPACKAGEUNIT}\n";
  $outstr .= "interface\n\n";
  $outstr .= "uses\n  Windows,\n  ActiveX;\n\n";
  $outstr .= "\n// \@REVIEW\@ (check for BCB directives, see proposal below!)\n\n";
  $outstr .= "// Proposal:\n// (*\$HPPEMIT '#include <$fname.h>' *)\n\n";
  # ACTUAL CONTENTS
  foreach(@{$arr})
  {
    $outstr .= $_;
  }
  # IMPLEMENTATION PART
  $outstr .= "\nimplementation\n\nend.\n";
  open(FILE, '>'.$fname.".pas");
  print FILE $outstr;
  close(FILE);
}

#------------------------------------------------------------------------------
# Get current date and time
#------------------------------------------------------------------------------
sub GetCurrentDateTime
{
  my ($sec, $min, $hour, $day, $month, $year, $dayofweek, $dayofyear, $dst) = localtime(time);
  $month+=1;
  $day = $day < 10 ? $day = "0".$day : $day;
  $month = $month < 10 ? $month = "0".$month : $month;
  $hour = $hour < 10 ? $hour = "0".$hour : $hour;
  $min = $min < 10 ? $min = "0".$min : $min;
  $sec = $sec < 10 ? $sec = "0".$sec : $sec;
  $year+=1900;
  return "$year-$month-$day\@$hour:$min:$sec";
}
__END__
This script has only one use: to convert an IDL (Interface Definition Language)
file to a Delphi unit
Since IDL is well-defined and less complex compared to C/C++ it's relatively
easy to convert an IDL file. The efforts have been solely done to serve the
JEDI project.

This script is mainly RX-driven (RX == regular expression).

Note, that the script may fail to catch all expressions. In this case you will
find the original lines (or in case of quoted C - the C lines) as comments with
the tag @REVIEW@ - and if you are lucky, the script had left some comments as
well. This usually happens only for C-preprocessor directives which can be
quite ambigous out of context ;-)
You MUST review the output file and check for the respective @REVIEW@ tag
and if necessary correct the part in question.

I wish you good luck and fast conversions with this little tool.

Oliver aka Assarbad
